{  _  )(_  _)(_  _)( ___) Object-oriented
 )(_)(   )(    )(   )__)  Tengwar
(_____) (__)  (__) (____) Transcription
version 0.0.9             Engine

a GNU-GPL software by <miszka@limes.com.pl>

OTTE is a library that provides you with an easy
way to create a transcriber from Latin alphabet
to Tolkien's Tengwar, like TengScribe or YaTT,
in Delphi or FPC. Although OTTE's transcription
back from Tengwar to Latin could be much better
for modes that place tehtar over the following
consonant (for ex. English or Sindarin Classic),
OTTE is even faster than Narmacil, and still in
development, so there may be new things added.
For now, see the program TENG.
-----------------------------------------------}

{$I otte_cfg.pas}
unit otte;

interface

uses
  otte_ctype,
  otte_math;

const
  CotteErrSetLocFailed = -1;
  CotteErrModNotOpened = -2;
  CotteErrModReadFault = -3;
  CotteErrModTooLarge  = -4;
  CotteErrModBadFormat = -5;
  CotteErrNotEnoughRAM = -6;
  CotteErrModNotLoaded = -7;
  
  CotteCharTransfAsRoman = true;
  CotteCharTransfAsTengwar = false;

  CotteTengwarQuestionMark = #174;
  CotteRomanQuestionMark = '?';
  CotteNoQuestionMark = #00;

  CotteModStringLen = 14;
  CotteModClauseSize = 36;
  CotteModDescLen = CotteModClauseSize-2;
  CotteModMaxClauses = 1500;

  CotteModAllChar = #00;
  CotteModNoAlpha = #13;

type
  TotteCharTransfCharSet = array[char] of boolean;
  TotteCharTransfIsNotAlpha = array[boolean] of TotteCharTransfCharSet;
  TotteCharTransfToLower = array[char] of char;

  TotteCharTransf = class(TObject)
    public
      IsNotAlpha: TotteCharTransfIsNotAlpha;
      ToLower: TotteCharTransfToLower;
      ToAdjustTehtar: TotteCharTransfToLower;
      RomanQuestionMark: char;
      TengwarQuestionMark: char;
      procedure reset;
      constructor create;
      function SetLocale(const locname: ansistring): integer;
      function StrLower(const s: ansistring): ansistring;
      function AdjustTehtar(const s: ansistring): ansistring;
  end;

  TotteEncodedModClause = array[0..CotteModClauseSize-1] of byte;
  TotteEncodedModClauses = array of TotteEncodedModClause;

  TotteModDesc = string[CotteModDescLen];
  TotteModString = string[CotteModStringLen];

  TotteModClause = record
    RomanMod, TengwarMod: TotteModString;
    NextRoman, LastTengwa, SetLastTengwa: char;
  end;

  TotteModClauses = array of TotteModClause;
  TotteModSortControl = function(const a, b: TotteModClause): boolean;

  TotteMode = class(TObject)
    protected
      RawMode: TotteEncodedModClauses;
    public
      CharTransf: TotteCharTransf;    //first set locale, then load mode
      clauses, BackClauses: TotteModClauses;
      description: TotteModDesc;
      CaseSensitive: boolean;
      procedure reset;
      constructor create;
      destructor destroy; override;
      procedure sort(var clauses: TotteModClauses; const control: TotteModSortControl);
      function load(const fname: ansistring): integer;
      function reload: integer;
    //function selfcheck(): boolean; //check if there are tabs...
    //function save(const fname: ansistring): integer; //still to write
      procedure AutoBackClauses;
  end;

  TotteCharIndex = array[char] of integer;

  TotteTransEngine = class(TObject)
    protected
      FirstInd, LastInd,
      FirstIndBack, LastIndBack: TotteCharIndex;
      IsWhiteSpace: array[char] of boolean;
      DoTranscription: boolean;
    public
      RevNum: boolean;        //reverse digit order
      CirNum: boolean;        //mark duedecades with circles beneath
      DotNum: boolean;        //mark numerals with dots
      BarNum: boolean;        //mark numerals with bars
      B12Num: boolean;        //use duodecimal numeric base
      TxtAll: boolean;        //don't treat numbers or spaces separately
      mode: TotteMode;
      procedure reset;
      constructor create;
      destructor destroy; override;
      function GetReady: integer;  //run this before any transcription task
      function TranscribeLine(TranscribeBack: boolean; const s: ansistring): ansistring;
      procedure StopTranscription;
  end;

const
  CotteCharTransfDefaultIsNotTengwar: TotteCharTransfCharSet =
  ( true,  true,  true,  true,  true,  true,  true,  true,
    true,  true,  true,  true,  true,  true,  true,  true,
    true,  true,  true,  true,  true,  true,  true,  true,
    true,  true,  true,  true,  true,  true,  true,  true,
    true,  false, false, false, false, false, false, false,
    false, false, false, false, false, true,  false, false,
    false, false, false, false, false, false, false, false,
    false, false, false, false, false, true,  false, false,
    false, false, false, false, false, false, false, false,
    false, false, false, false, false, false, false, false,
    false, false, false, false, false, false, false, false,
    false, false, false, false, false, false, false, false,
    false, false, false, false, false, false, false, false,
    false, false, false, false, false, false, false, false,
    false, false, false, false, false, false, false, false,
    false, false, false, false, false, false, false, true,
    true,  true,  false, false, false, false, false, false,
    true,  false, false, false, true,  true,  true,  true,
    true,  false, false, false, false, false, false, false,
    false, false, false, true,  true,  true,  true,  false,
    true,  false, false, false, true,  false, false, false,
    false, false, false, true,  true,  false, true,  false,
    false, true,  true,  false, false, false, true,  true,
    false, true,  false, true,  false, false, false, false,
    true,  true,  true,  true,  false, false, false, true,
    false, false, false, false, false, false, false, false,
    false, false, false, false, false, false, false, false,
    false, false, false, false, false, false, false, false,
    false, false, false, false, false, false, false, false,
    false, false, false, false, false, false, false, false,
    true,  true,  true,  true,  true,  true,  true,  true,
    true,  true,  true,  true,  false, false, false, false  );

  CotteIsNonSpacingTengwar: TotteCharTransfCharSet =
  ( false, false, false, false, false, false, false, false,
    false, false, false, false, false, false, false, false,
    false, false, false, false, false, false, false, false,
    false, false, false, false, false, false, false, false,
    false, false, false, true,  true,  true,  true,  true,
    true,  true,  false, false, false, false, false, true,
    true,  false, false, false, false, false, false, false,
    false, false, true,  true,  false, false, true,  true,
    false, false, true,  true,  true,  true,  true,  true,
    true,  false, true,  false, true,  true,  true,  true,
    true,  false, true,  false, true,  true,  true,  false,
    false, true,  false, true,  false, false, true,  false,
    false, false, false, false, false, false, false, false,
    false, false, false, false, false, false, false, false,
    true,  false, false, false, false, false, false, false,
    false, false, false, true,  true,  true,  false, false,
    false, false, true,  true,  true,  true,  true,  false,
    false, true,  true,  true,  false, false, false, false,
    false, true,  true,  true,  true,  false, false, false,
    true,  true,  false, false, false, false, false, true,
    true,  false, false, false, false, false, false, false,
    true,  true,  true,  false, false, true,  false, true,
    true,  false, false, false, true,  true,  false, false,
    true,  false, false, false, false, false, false, false,
    false, false, false, false, false, false, false, false,
    true,  true,  true,  true,  true,  true,  true,  true,
    true,  true,  true,  true,  true,  true,  true,  true,
    true,  true,  true,  true,  true,  true,  true,  true,
    true,  true,  true,  true,  true,  true,  true,  true,
    true,  true,  true,  true,  true,  true,  true,  true,
    false, false, false, false, false, false, false, false,
    false, false, false, false, true,  true,  true,  true   );

  CotteCharTransfAdjustTehtar: TotteCharTransfToLower =
  ( #0,    #1,    #2,    #3,    #4,    #5,    #6,    #7,
    #8,    #9,    #10,   #11,   #12,   #13,   #14,   #15,
    #16,   #17,   #18,   #19,   #20,   #21,   #22,   #23,
    #24,   #25,   #26,   #27,   #28,   #29,   #30,   #31,
    #32,   #33,   #39,   #35,   #36,   #37,   #38,   #39,
    #40,   #48,   #42,   #56,   #44,   #45,   #46,   #59,
    #48,   #49,   #50,   #51,   #52,   #53,   #54,   #55,
    #56,   #57,   #59,   #59,   #60,   #61,   #62,   #59,
    #64,   #65,   #37,   #35,   #35,   #35,   #36,   #37,
    #94,   #56,   #38,   #75,   #76,   #38,   #94,   #79,
    #48,   #81,   #36,   #83,   #37,   #38,   #36,   #87,
    #88,   #94,   #90,   #91,   #92,   #93,   #94,   #56,
    #96,   #97,   #98,   #99,   #100,  #101,  #102,  #103,
    #104,  #105,  #106,  #107,  #108,  #109,  #110,  #111,
    #48,   #113,  #114,  #115,  #116,  #117,  #118,  #119,
    #120,  #121,  #122,  #91,   #56,   #56,   #126,  #127,
    #128,  #129,  #130,  #131,  #131,  #131,  #131,  #135,
    #136,  #137,  #138,  #139,  #140,  #141,  #142,  #143,
    #144,  #145,  #145,  #145,  #145,  #149,  #150,  #151,
    #152,  #153,  #154,  #155,  #156,  #157,  #158,  #159,
    #160,  #56,   #56,   #56,   #164,  #56,   #166,  #167,
    #168,  #169,  #170,  #171,  #172,  #170,  #174,  #170,
    #176,  #177,  #178,  #179,  #180,  #170,  #182,  #183,
    #184,  #185,  #186,  #187,  #188,  #189,  #190,  #191,
    #192,  #193,  #194,  #195,  #196,  #56,   #56,   #199,
    #200,  #201,  #202,  #203,  #204,  #205,  #206,  #207,
    #62,   #62,   #62,   #62,   #212,  #212,  #212,  #212,
    #216,  #216,  #216,  #216,  #220,  #220,  #220,  #220,
    #224,  #224,  #224,  #224,  #228,  #228,  #228,  #228,
    #232,  #232,  #232,  #232,  #91,   #39,   #91,   #39,
    #240,  #241,  #242,  #243,  #244,  #245,  #246,  #247,
    #248,  #249,  #250,  #251,  #252,  #252,  #252,  #252 );

implementation
//===TotteCharTransf===================================================================
procedure TotteCharTransf.reset;
begin
  self.IsNotAlpha[CotteCharTransfAsTengwar] := CotteCharTransfDefaultIsNotTengwar;
  self.RomanQuestionMark := CotteRomanQuestionMark;
  self.TengwarQuestionMark := CotteTengwarQuestionMark;
  self.ToAdjustTehtar := CotteCharTransfAdjustTehtar;
  self.SetLocale('');
end;

//-----------------------------------------------------------------------------
constructor TotteCharTransf.create;
begin
  inherited;
  self.reset;
end;

//-----------------------------------------------------------------------------
function TotteCharTransf.StrLower(const s: ansistring): ansistring;

var
  i: integer;

begin
  result := s;
  for i := 1 to length(s) do result[i] := ToLower[s[i]];
end;

//-----------------------------------------------------------------------------
function TotteCharTransf.AdjustTehtar(const s: ansistring): ansistring;

var
  i: integer;

begin
  result := s;
  for i := 1 to length(s) do result[i] := ToAdjustTehtar[s[i]];
end;

//-----------------------------------------------------------------------------
function TotteCharTransf.SetLocale(const locname: ansistring): integer;

var
  c: char;

begin
  if otte_ctype.setlocale(locname) then begin
    result := 0;
    
    for c := #0 to #255 do begin
      self.IsNotAlpha[CotteCharTransfAsRoman, c] := not otte_ctype.isalpha(c);
      self.ToLower[c] := otte_ctype.tolower(c);
    end;
  end
  else result := CotteErrSetLocFailed;
end;

//===TotteMode==================================================================
procedure TotteMode.reset;
begin
  CharTransf.reset;
  SetLength(clauses, 0);
  SetLength(BackClauses, 0);
  SetLength(RawMode, 0);  
  description := '';
  CaseSensitive := false;
end;

//-----------------------------------------------------------------------------
constructor TotteMode.create;
begin
  inherited;
  CharTransf := TotteCharTransf.create;
  reset;
end;

//-----------------------------------------------------------------------------
destructor TotteMode.destroy;
begin
  reset;
  CharTransf.destroy;
  inherited;
end;

//-----------------------------------------------------------------------------
procedure TotteMode.sort(var clauses: TotteModClauses; const control: TotteModSortControl);

var
  i, j: integer;
  t: TotteModClause;

{ Do NOT use any unstable sorting method!
  Also, Insertionsort is fast enough.
}

begin
  if Addr(control) = nil then exit
  else for i := 1 to high(clauses) do begin
    j := i;
    t := clauses[j];

    while (j>0) and control(clauses[j-1], t) do begin
      clauses[j] := clauses[j-1];
      dec(j);
    end;

    clauses[j] := t;
  end;
end;

//-----------------------------------------------------------------------------
function TotteMode.load(const fname: ansistring): integer;

var
  f: file;
  fsize, count: integer;
  data, swap: TotteEncodedModClauses;
  ClausesCount: integer;

begin
  result := 0;
  setlength(swap, 0);
  setlength(data, 0);
  
  assign(f, fname);
  system.reset(f, 1);

  if IOResult<>0 then begin
    result := CotteErrModNotOpened;
    exit;
  end;
  
  fsize := FileSize(f);
  ClausesCount := fsize div CotteModClauseSize - 2;

  if ((fsize mod CotteModClauseSize)<>0) or (ClausesCount<=0) then 
    result := CotteErrModBadFormat
  else if ClausesCount > CotteModMaxClauses then
    result := CotteErrModTooLarge;

  if result<>0 then begin
    close(f);
    exit;
  end;

  try
    SetLength(data, ClausesCount+2);
  except
    { Code below gets repeated and repeated.
      I must rewrite it once.
    }
    result := CotteErrNotEnoughRAM;
    close(f);
    SetLength(data, 0);
    exit;
  end;

  BlockRead(f, data[0], fsize, count);

  if (IOResult<>0) or (count<>fsize) then begin
    result := CotteErrModReadFault;
    close(f);
    SetLength(data, 0);
    exit;
  end;

  close(f);

  swap := self.RawMode;
  self.RawMode := data;
  result := reload;
  if result <> 0 then self.RawMode := swap;
end;

//-----------------------------------------------------------------------------
function TotteMode.reload: integer;

  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
  function chop(var s: ansistring): ansistring;
  begin
    result := '';

    while (s<>'') and (s[1]<>#09) do begin
      result := result + s[1];
      delete(s, 1, 1);
    end;

    if s<>'' then delete(s, 1, 1);
  end;
  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
  function DecodeHeader(
    const data: TotteEncodedModClause;
    var Description: TotteModDesc;
    var CaseSensitive: boolean
  ): integer;

  var
    i: integer;
    s, desc, csflag: ansistring;

  begin
    result := 0;

    s := '';
    for i := 0 to CotteModClauseSize-1 do s := s + char(data[i]);

    desc := chop(s);
    if length(desc) > CotteModDescLen then begin
      result := CotteErrModBadFormat;
      exit;
    end;

    csflag := chop(s);
    if (csflag<>'0') and (csflag<>'1') then begin
      result := CotteErrModBadFormat;
      exit;
    end;

    Description := desc;
    if csflag = '0' then CaseSensitive := false
    else if csflag = '1' then CaseSensitive := true
  end;
  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
  function DecodeClause(
    const data: TotteEncodedModClause;
    var clause:  TotteModClause
  ): integer;

  var
    i: integer;
    s, rommod, tengmod, nextr, lastt, setlastt: ansistring;

  begin
    result := 0;

    s := '';
    for i := 0 to CotteModClauseSize-1 do s := s + char(data[i]);    

    rommod := chop(s);
    if (rommod='') or (length(rommod)>CotteModStringLen) then begin
      result := CotteErrModBadFormat;
      exit;
    end;
   
    tengmod := chop(s);
    if (length(tengmod)>CotteModStringLen) then begin
      result := CotteErrModBadFormat;
      exit;
    end;
    
    nextr := chop(s);
    if (nextr='') then nextr := CotteModAllChar;

    lastt := chop(s);
    if (lastt='') then lastt := CotteModAllChar;
    
    setlastt := chop(s);
    if (setlastt='') then setlastt := CotteModAllChar;

    with clause do begin
      RomanMod := rommod;
      TengwarMod := tengmod;
      NextRoman := nextr[1];
      LastTengwa := lastt[1];
      SetLastTengwa := setlastt[1];
      
      if self.CaseSensitive then begin
        RomanMod := self.CharTransf.StrLower(RomanMod);
	NextRoman := self.CharTransf.ToLower[NextRoman];
      end;
    end;
  end;
  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

var
  TempClauses: TotteModClauses;
  desc: TotteModDesc;
  csflag: boolean;
  i: integer;

begin
  if length(self.RawMode) < 3 then result := CotteErrModNotLoaded
  else begin
    try
      SetLength(TempClauses, length(self.RawMode)-2);
    except
      result := CotteErrNotEnoughRAM;
      SetLength(TempClauses, 0);
      exit;
    end;

    result := DecodeHeader(self.RawMode[0], desc, csflag);

    if result<>0 then begin
      SetLength(TempClauses, 0);
      exit;
    end;

    for i := 2 to High(self.RawMode) do begin
      result := DecodeClause(self.RawMode[i], TempClauses[i-2]);

      if result<>0 then begin
        SetLength(TempClauses, 0);
        exit;
      end;
    end;

    if result=0 then begin
      self.CaseSensitive := csflag;
      self.description := desc;
      self.clauses := TempClauses;
      self.AutoBackClauses;
      {
      Don't finalize TempClauses!
      That will erase our new self.clauses!
      For similar, don't finalize old self.clauses!
      }
    end
    else SetLength(TempClauses, 0);
  end;
end;

//-----------------------------------------------------------------------------
function SortBackClauses(const a, b: TotteModClause): boolean; register;
begin
  if a.RomanMod[1] = b.RomanMod[1] then begin
    if a.RomanMod[0] = b.RomanMod[0] then begin
      if a.RomanMod = b.RomanMod then result := a.TengwarMod[0] > b.TengwarMod[0]
      else result := a.RomanMod > b.RomanMod;
    end
    else result := a.RomanMod[0] < b.RomanMod[0];
  end
  else result := a.RomanMod[1] > b.RomanMod[1];
end;

//-----------------------------------------------------------------------------
// This procedure attempts to guess how to transcribe the tengwar text produced
// with the current mode file back to Roman. The result is stored in BackClauses,
// in the same way as normal mode file. Note: types of Clauses and BackClauses
// are identical, but fields of BackClauses have "inverted" meaning.
procedure TotteMode.AutoBackClauses;

var
  TempClauses: TotteModClauses;
  TempStr: TotteModString;
  ss: TotteModSortControl;
  FindByTengwa: array[char] of char;
  Found: boolean;
  i, j: integer;

  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
  //This function is used to check if the first clause overlaps with the second.
  function ClausesEqual(const clause1, clause2: TotteModClause): boolean;
  begin
    result :=
      (clause1.RomanMod = clause2.RomanMod) and
      (clause1.TengwarMod = clause2.TengwarMod) and
      (clause1.NextRoman = clause2.NextRoman) and
      (clause1.LastTengwa = clause2.LastTengwa) and
      (clause1.SetLastTengwa = clause2.SetLastTengwa);
  end;
  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

begin
  //Allocate a temporary workplace
  setlength(TempClauses, length(self.clauses));
  //set initial FindByTengwa entries
  fillchar(FindByTengwa, sizeof(FindByTengwa), byte(CotteModAllChar));
  For i := 1 to 32 do FindByTengwa[char(i)] := char(i);

  //Copy and preprocess all significant records to a temporary location
  j := 0;
  for i := 0 to high(self.clauses) do
  if self.clauses[i].TengwarMod <> '' then begin
    TempClauses[j] := self.clauses[i];
    inc(j);
  end;
  setlength(TempClauses, j);

  //Do the first phase of translation
  for i := 0 to high(TempClauses) do
  with TempClauses[i] do begin
    //Establish a Tengwa-Roman relationship
    if (RomanMod[0] = #1) and (TengwarMod[0] = #1) and
       (NextRoman = CotteModAllChar) and (LastTengwa = CotteModAllChar) and
       (FindByTengwa[TengwarMod[1]] = CotteModAllChar) then
    FindByTengwa[TengwarMod[1]] := RomanMod[1];

    //Try harder
    if (RomanMod[0] = #1) and (TengwarMod[0] = #1) and
       (FindByTengwa[TengwarMod[1]] = CotteModAllChar) then
    FindByTengwa[TengwarMod[1]] := RomanMod[1];

    //Also, swap ROMAN with TENGWAR
    TempStr := RomanMod;
    RomanMod := self.CharTransf.AdjustTehtar(TengwarMod);
    TengwarMod := TempStr;
  end;  

  //Now it's time to process the records
  for i := 0 to High(TempClauses) do begin
    with TempClauses[i] do begin

      //see some most common cases
      if (LastTengwa = CotteModAllChar) and
	       (NextRoman = CotteModAllChar) then
      begin
        if (SetLastTengwa = RomanMod) and (TengwarMod[0] = #1) then
          SetLastTengwa := TengwarMod[1]
      	else SetLastTengwa := FindByTengwa[SetLastTengwa];
      end
      else if 
         (length(TengwarMod) >= 1) and
         (LastTengwa = SetLastTengwa) and
	 (NextRoman in [CotteModAllChar, CotteModNoAlpha]) then
      begin
        LastTengwa := FindByTengwa[LastTengwa];
        SetLastTengwa := LastTengwa;
      end
      else if
         (length(TengwarMod) = 1) and
	 (RomanMod = SetLastTengwa) and
	 (NextRoman = CotteModAllChar) then
      begin
	LastTengwa := CotteModAllChar;
	SetLastTengwa := TengwarMod[1];
      end
      else begin //Handle less common cases...

        //translate NEXT ROMAN to NEXT TENGWA
        Found := false;

        if not (NextRoman in [CotteModAllChar, CotteModNoAlpha]) then begin
          for j := 0 to High(self.clauses) do begin
            if (self.clauses[j].TengwarMod = NextRoman) then begin
              NextRoman := self.clauses[j].RomanMod[1];

              //if vowel carrier, then take the next char
              if (NextRoman in ['`', '~']) and (Length(self.clauses[j].TengwarMod)>1) then
                NextRoman := self.clauses[j].TengwarMod[2];

              NextRoman := CharTransf.ToAdjustTehtar[NextRoman];
              Found := true;
              break;
            end;
          end;
        end;

        if not Found then NextRoman := CotteModAllChar;

        //translate LAST TENGWA to LAST ROMAN
        LastTengwa := FindByTengwa[LastTengwa];

        //translate SET LAST TENGWA to SET LAST ROMAN
        SetLastTengwa := FindByTengwa[SetLastTengwa];
      end;
    end;
  end;

  //sort clauses
  ss := SortBackClauses;
  self.sort(TempClauses, ss);

  //remove subsequent overlapping clauses
  i := 1;
  while i <= high(TempClauses) do begin
    if ClausesEqual(TempClauses[i-1], TempClauses[i]) then begin
      for j := i to high(TempClauses) do TempClauses[j-1] := TempClauses[j];
      SetLength(TempClauses, length(TempClauses)-1);
    end
    else inc(i);
  end;

  //remove overlapping clauses

  //store result
  self.BackClauses := TempClauses;
end;

//===TotteTransEngine==========================================================
procedure TotteTransEngine.reset;
begin
  mode.reset;
  DoTranscription := true;
  RevNum := false;
  CirNum := false;
  DotNum := false;
  BarNum := false;
  B12Num := false;
  TxtAll := false;
end;

//-----------------------------------------------------------------------------
constructor TotteTransEngine.create;
begin
  inherited;
  mode := TotteMode.create;
  reset;
end;

//-----------------------------------------------------------------------------
destructor TotteTransEngine.destroy;
begin
  mode.destroy;
  inherited;
end;

//-----------------------------------------------------------------------------
function SortControl(const a, b: TotteModClause): boolean; register;
begin
  result := a.RomanMod[1] > b.RomanMod[1];
end;

//-----------------------------------------------------------------------------
function TotteTransEngine.GetReady: integer;
{
  Run this proc whenever you have just set all the params needed to do
  the transcription and before you start. If you don't, you'll get
  erroneous results!

  This proc prepares some data for transcription process. It also sorts
  and indexes the mode clauses by their first leotters to speed the
  transcription process up (about 1.5 times).
}
  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
  function GetHalfReady(var Clauses: TotteModClauses;
                        var FirstInd, LastInd: TotteCharIndex): integer;
  var
    c: char;
    i: integer;

  begin
    if length(clauses) = 0 then begin
      result := CotteErrModNotLoaded;
      exit;
    end
    else result := 0;

    //Sort the mode clauses
    mode.sort(clauses, SortControl);

    //Index the mode clauses by the first char of the ROMAN.
    //Chars that do not open any mode rule are given strange
    //indexes to be immediately skipped during thanscription
    for c := #00 to #255 do begin
      FirstInd[c] := high(clauses);

      for i := 0 to high(clauses) do begin
        if clauses[i].RomanMod[1] = c then begin
          FirstInd[c] := i;
          break;
        end;
      end;

      LastInd[c] := 0;

      for i := high(clauses) downto 0 do begin
        if clauses[i].RomanMod[1] = c then begin
          LastInd[c] := i;
          break;
        end;
      end;
    end;
  end;
  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

begin
  //Set characters skipped in transcription process
  fillchar(IsWhiteSpace, SizeOf(IsWhiteSpace), false);
  fillchar(IsWhiteSpace, 32, true);

  if not TxtAll then begin
    IsWhiteSpace[#32] := true;
    IsWhiteSpace[#160] := true;
  end;

  result := GetHalfReady(mode.clauses, FirstInd, LastInd);
  if result = 0 then
    result := GetHalfReady(mode.BackClauses, FirstIndBack, LastIndBack);
end;

//--DEVICES-FOR-TRANSCRIBELINE-------------------------------------------------
{ These are functions for determining whether the following char in the
  buffer is a digit or not, IsDigit variable points to the suitable one. }

type TotteIsDigit = function(c: char): boolean;

//. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
function IsRomanDigit(c: char): boolean;
begin
  result := c in ['0'..'9'];
end;

//. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
function IsTengwarDigit(c: char): boolean;
begin
  result := c in [#240..#251];
end;

{ These are functions for cutting numeral from the buffer.
  CutNumeral variable points to the suitable one. }

type TotteCutNumeral = function(const buffer: ansistring; var index: integer): ansistring;

//. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
function CutRomanNumeral(const buffer: ansistring; var index: integer): ansistring;
var count: integer;
begin
  count := 0;
  while (index+count <= length(buffer)) and IsRomanDigit(buffer[index+count]) do inc(count);
  Result := Copy(buffer, index, count);
  inc(index, count);    
end;

//. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
function CutTengwarNumeral(const buffer: ansistring; var index: integer): ansistring;
var count: integer;
begin
  count := 0;
  result := '';
  while (index+count <= length(buffer)) and
        (IsTengwarDigit(buffer[index+count]) or
	 CotteIsNonSpacingTengwar[buffer[index+count]]) do
  begin
    if IsTengwarDigit(buffer[index+count]) then
      result := result + buffer[index+count];

    inc(count);
  end;
  inc(index, count);    
end;

//-----------------------------------------------------------------------------
function TotteTransEngine.TranscribeLine(TranscribeBack: boolean; const s: ansistring): ansistring;

var
  buffer: ansistring;
  LastWrittenTengwa: char;
  index: integer;
  {$IFDEF WINDOWS}ousindex: integer;{$ENDIF}

  //Either points to mode.clauses or mode.BackClauses
  clauses: TotteModClauses;

  //Either pointers to FirstRomanInd/LastRomanInd
  //or to FirstTengwarInd/LastTengwarInd
  FirstIndex, LastIndex: ^TotteCharIndex;

  //Either mode.CharTransf.RomanQuestionMark
  //or mode.CharTransf.TengwarQuestionMark
  QuestionMark: char;

  //Indexes to mode.CharTransf.IsNotAlpha[boolean].
  //Swapped when transcribing back: IsNotAlpha[AsRoman]
  //returns IsNotTengwar then, and thus these indexes
  //are separated from global otte.CotteCharTransfAsRoman
  //and otte.CotteCharTransfAsTengwar.
  AsRoman, AsTengwar: boolean;

  //Either b10_to_b12 or b12_to_b10;
  BaseConv: TotteBaseConv;

  //Either IsRomanDigit or IsTengwarDigit;
  IsDigit: TotteIsDigit;

  //Either CutRomanNumeral or CutTengwarNumeral;
  CutNumeral: TotteCutNumeral;

  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
  {$IFDEF WINDOWS}
  { This is a workaround for poor windows memory management that slows down
    standard string concatenation. It explicitly pre-reserves space for output
    string and calls "move". No out-of-memory checking, sorry!
  }
  procedure StrAdd(const ins: ansistring; var ousindex: Integer; var ous: ansistring);

  var
    count: integer;

  begin
    count := length(ins);
    if count <> 0 then begin
      if (ousindex+count > length(ous)) then SetLength(ous, ousindex+count+80);
      move(ins[1], ous[ousindex], count);
      inc(ousindex, count);
    end;
  end;
  {$ENDIF}

  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
  function ProcessNumerals(const s: ansistring): ansistring;

  const
    //indexes for Nums[digit, index] array
    //Note for barover and barunder: they are taken in account only for first
    //written digit, as the bar  may be longer or shorter. Then, for the second
    //and following digits, the standard, long bars are used.
    numeral  = 0;
    dotover  = 1;
    barover  = 2;
    dotunder = 3;
    barunder = 4;
    circle   = 5;

    // standard bars
    LongBarOver  = #236;  //over second and further digits
    LongBarUnder = #237;  //under second and further digits

    // Array of Elvish numerals and their "attributes"
    Nums: array[0..11, numeral..circle] of char =

     // numeral dotover barover dotunder barunder circle      digit
        ((#240,   'T',   #238,   #201,    #239,   #168),     {  0  }
         (#241,   'G',   #238,   #202,    #239,   #168),     {  1  }
         (#242,   '%',   #236,   #200,    #237,   #153),     {  2  }
         (#243,   '%',   #236,   #200,    #237,   #152),     {  3  }
         (#244,   'G',   #238,   #202,    #239,   #153),     {  4  }
         (#245,   'T',   #238,   #201,    #239,   #152),     {  5  }
         (#246,   '%',   #236,   #200,    #237,   #152),     {  6  }
         (#247,   'T',   #238,   #201,    #239,   #153),     {  7  }
         (#248,   'G',   #238,   #202,    #239,   #153),     {  8  }
         (#249,   '%',   #236,   #200,    #237,   #152),     {  9  }
         (#250,   'B',   #238,   #202,    #239,   #153),     {  10 }
         (#251,   'B',   #238,   #202,    #239,   #153));    {  11 }

  var
    NumBuff: ansistring;    //ASCII string that holds the number
    i: integer;
    k: integer;             //we write NumBuff[k], not NumBuff[i]
    NumLen: Integer;        //length of input numeral
    j: Byte;                //NumBuff[k] converted into index of Nums array
    c: char;

  begin
    NumBuff := s;
    Result := '';

    //revert digit order BEFORE processing if back transcription
    NumLen := Length(NumBuff);
    if TranscribeBack and RevNum then
    for i := 1 to length(NumBuff) div 2 do begin
      c := NumBuff[i];
      NumBuff[i] := NumBuff[NumLen - i + 1];
      NumBuff[NumLen - i + 1] := c;
    end;

    //convert into duodecimal, if requested.
    if B12Num then NumBuff := BaseConv(NumBuff);

    NumLen := Length(NumBuff);

    //write the numeral now
    if TranscribeBack then
      for i := 1 to NumLen do
        Result := Result + char(byte(NumBuff[i]) - 240 + byte('0'))
    else for i := 1 to NumLen do begin
      if RevNum then k := NumLen-i+1 else k := i; //do you want reverse order?

      //write the digit
      j := Byte( NumBuff[k] ) - Ord('0');
      Result := Result + Nums[j, numeral];

      //if dots requested
      if DotNum then begin
        if B12Num then begin
          if (not CirNum) or (k <> NumLen-1) then
            Result := Result + Nums[j, dotunder];
        end
        else Result := Result + Nums[j, dotover];
      end;

      //if bars requested
      if BarNum then begin
        if (k=1) then begin //first wriotten digit - use sophisticated bars
          if B12Num then Result := Result + Nums[j, barunder]
          else Result := Result + Nums[j, barover];
        end
        else begin //use standard long bars
          if B12Num then Result := Result + LongBarUnder
          else Result := Result + LongBarOver;
        end;
      end;

      //if  circled duodecades requested
      if CirNum and (k = NumLen-1) then Result := Result + (Nums[j, circle]);
    end;
  end;
  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
  function CharsMatch(ch, ch_mod: char; AsRoman: boolean): boolean;
  begin
    Result := (ch_mod = CotteModAllChar) or (ch = ch_mod) or
              ((ch_mod = CotteModNoAlpha) and mode.CharTransf.IsNotAlpha[AsRoman, ch]);
  end;
  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
  function ProcessText(const buffer: ansistring; var index: integer): TotteModString;

  var
    i: Integer;
    Temp: string[CotteModStringLen+1];
    modclause: ^TotteModClause;

  begin
    Result := '';
    Temp := '               ';
    Temp := copy(buffer, index, CotteModStringLen+2);
    modclause := addr(clauses[FirstIndex[Temp[1]]]);
    Dec(modclause); //don't dereference this now, may be out of range!

    //here GetReady indexing efforts take effect
    //we don't have to search through all the mode clauses
    for i := FirstIndex[Temp[1]] to LastIndex[Temp[1]] do begin
      Inc(modclause); //now it is surely in the range

      with modclause^ do begin
        Temp[0] := RomanMod[0];

        //further speed improvement - great for complicated modes (i.e. Quenya)
	//caution - checking characters no. 4 and next slows things down...
	//takes no effect for simple modes (i. e. Sindarin_beleriand)
        if (Temp[0] >= #2) and (RomanMod[2] <> Temp[2]) then continue
        else if (Temp[0] >= #3) and (RomanMod[3] <> Temp[3]) then continue
        else if
           CharsMatch(LastWrittenTengwa, LastTengwa, AsTengwar) and
           CharsMatch(Temp[Byte(Temp[0])+1], NextRoman, AsRoman) and
	   ((Temp[0] <= #3) or (RomanMod = Temp)) then
        begin
          result := TengwarMod;
          LastWrittenTengwa := SetLastTengwa;
	  inc(index, length(RomanMod));
          exit;
        end;
      end;
    end;

    //when transcription error...
    if QuestionMark <> CotteNoQuestionMark then begin
      result := QuestionMark;
      LastWrittenTengwa := ' ';
    end;
    
    inc(index);
  end;
  //. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

begin
  if TranscribeBack then begin
    IsDigit := IsTengwarDigit;
    CutNumeral := CutTengwarNumeral;
    BaseConv := b12_to_b10;

    clauses := self.mode.BackClauses;
    QuestionMark := self.mode.CharTransf.RomanQuestionMark;
    AsRoman := CotteCharTransfAsTengwar; //swapped
    AsTengwar := CotteCharTransfAsRoman; //swapped
    FirstIndex := @FirstIndBack;
    LastIndex := @LastIndBack;

    buffer := self.mode.CharTransf.AdjustTehtar(s);
  end
  else begin
    IsDigit := IsRomanDigit;
    CutNumeral := CutRomanNumeral;
    BaseConv := b10_to_b12;

    clauses := self.mode.Clauses;
    QuestionMark := self.mode.CharTransf.TengwarQuestionMark;
    AsRoman := CotteCharTransfAsRoman;
    AsTengwar := CotteCharTransfAsTengwar;
    FirstIndex := @FirstInd;
    LastIndex := @LastInd;

    if mode.CaseSensitive then buffer := s
    else buffer := self.mode.CharTransf.StrLower(s);
  end;

  result := '';
  if length(mode.clauses) = 0 then exit;
  index := 1;
  
  {$IFDEF WINDOWS}
  ousindex := 1;
  setLength(result, length(s));
  {$ENDIF}
  
  LastWrittenTengwa := ' ';

  //if transcription was cancelled before
  if not DoTranscription then DoTranscription := true;

  while (index <= length(buffer)) and DoTranscription do begin
    if IsWhiteSpace[buffer[index]] then begin
      {$IFDEF WINDOWS}StrAdd(buffer[index], ousindex, result);
      {$ELSE}result := result + buffer[index];
      {$ENDIF}
      inc(index);
      LastWrittenTengwa := ' ';
    end
    else if (not TxtAll) and IsDigit(buffer[index]) then begin
      {$IFDEF WINDOWS}StrAdd(ProcessNumerals(CutNumeral(buffer, index)), ousindex, result);
      {$ELSE}result := result + ProcessNumerals(CutNumeral(buffer, index));
      {$ENDIF}
      LastWrittenTengwa := CotteModNoAlpha;
    end
    else begin
      {$IFDEF WINDOWS}StrAdd(ProcessText(buffer, index), ousindex, result);
      {$ELSE}result := result + ProcessText(buffer, index);
      {$ENDIF}
    end;
  end;
  
  {$IFDEF WINDOWS}SetLength(result, ousindex-1);{$ENDIF}
  
  //if transcription was cancelled in progress
  if not DoTranscription then DoTranscription := true;
end;

//-----------------------------------------------------------------------------
procedure TotteTransEngine.StopTranscription;
begin
  DoTranscription := false;
end;

//===EndOfImplementation=======================================================
end.
